@@ -20,6 +20,7 @@ match:
 		const void *ip,
 		const struct xt_match *match,
 		void *matchinfo,
+		unsigned int matchinfosize,
 		unsigned int hook_mask,
 	);
 
@@ -27,6 +28,7 @@ match:
 	(*destroy)(
 		const struct xt_match *match,
 		void *matchinfo,
+		unsigned int matchinfosize,
 	);
 
 target:
@@ -40,6 +42,7 @@ target:
 		unsigned int hooknum,
 		const struct xt_target *target,
 		const void *targinfo,
+		void *userdata,
 	);
 
 	/* true/false */
@@ -49,6 +52,7 @@ target:
 		const void *entry,
 		const struct xt_target *target,
 		void *targinfo,
+		unsigned int targinfosize,
 		unsigned int hook_mask,
 	);
 
@@ -56,4 +60,5 @@ target:
 	(*destroy)(
 		const struct xt_target *target,
 		void *targinfo,
+		unsigned int targinfosize,
 	);
